#!/usr/bin/env python3

import argparse
import sys
import time
import logging
import grpc
import os
import json
import math
from google.protobuf import text_format

from nuance.tts.storage.v1beta1.storage_pb2 import *
from nuance.tts.storage.v1beta1.storage_pb2_grpc import *

args = None

def parse_args():
    global args
    parser = argparse.ArgumentParser(
        prog="storage-client.py",
        usage="%(prog)s [-options]",
        add_help=False,
        formatter_class=lambda prog: argparse.HelpFormatter(
            prog, max_help_position=45, width=100)
    )

    options = parser.add_argument_group("options")
    options.add_argument("-h", "--help", action="help",
                         help="Show this help message and exit")
    options.add_argument("--server_url", nargs="?",
                         help="Server hostname (default=localhost)", default="localhost:8080")
    options.add_argument("--token", nargs="?",
                         help="Access token", required=True)
    options.add_argument("--max_chunk_size_bytes", metavar="num", nargs="?",
                         help="Max size (in bytes) of each file chunk, default=4096", default=4096, type=int)
    options.add_argument("--upload", action="store_true",
                         help="Send an Upload RPC. Requires the --context_tag, --name and other resource-specific options.")
    options.add_argument("--delete", action="store_true",
                         help="Send a Delete RPC. Requires the --uri option.")
    options.add_argument("--file", metavar="file", nargs="?",
                         help="File to upload. If an ActivePrompt Database, must be packaged as a zip.")
    options.add_argument("--context_tag", metavar="tag", nargs="?",
                         help="Context tag", default='')
    options.add_argument("--name", metavar="name", nargs="?",
                         help="Resource name", default='')
    options.add_argument("--type", metavar="type", nargs="?",
                         help="Resource type. Must be one of: [activeprompt, user_dictionary, text_ruleset, wav]")
    options.add_argument("--language", metavar="type", nargs="?",
                         help="IETF language code. Required if type is [user_dictionary, text_ruleset])", default='')
    options.add_argument("--voice", metavar="type", nargs="?",
                         help="ActivePrompt voice", default='')
    options.add_argument("--voice_model", metavar="type", nargs="?",
                         help="ActivePrompt voice model", default='')
    options.add_argument("--voice_version", metavar="type", nargs="?",
                         help="ActivePrompt voice version", default='')
    options.add_argument("--vocalizer_studio_version", metavar="type", nargs="?",
                         help="ActivePrompt Vocalier Studio version", default='')
    options.add_argument("--uri", metavar="uri", nargs="?",
                         help="URI to delete", default='')
    args = parser.parse_args()

def create_channel():
    call_credentials = None
    channel = None

    call_credentials = grpc.access_token_call_credentials(args.token)
    channel_credentials = grpc.composite_channel_credentials(grpc.ssl_channel_credentials(), call_credentials)

    return grpc.secure_channel(args.server_url, credentials=channel_credentials)

# Reads the input file in chunks
def read_file(file, context_tag, name, type, voice, voice_model, voice_version, vocalizer_studio_version, language, max_chunk_size_bytes=4096):

    file_handle = open(file, "rb")

    upload_request = UploadRequest()
    upload_init_message = UploadInitMessage()
    upload_init_message.context_tag = args.context_tag
    upload_init_message.name = args.name

    if type == 'activeprompt':
        log.info('Type is ActivePromptDB')
        active_prompt_db = ActivePromptDB()
        active_prompt_db.voice = voice
        active_prompt_db.voice_model = voice_model
        active_prompt_db.voice_version = voice_version
        active_prompt_db.vocalizer_studio_version = vocalizer_studio_version
        upload_init_message.active_prompt_db.CopyFrom(active_prompt_db)
    elif type == "user_dictionary":
        log.info('Type is User Dictionary')
        user_dictionary = UserDictionary()
        user_dictionary.language = language
        upload_init_message.dictionary.CopyFrom(user_dictionary)
    elif type == "text_ruleset":
        log.info('Type is Text User Ruleset')
        text_ruleset = TextUserRuleset()
        text_ruleset.language = language
        upload_init_message.text_ruleset.CopyFrom(text_ruleset)
    elif type == "wav":
        log.info('Type is Wav')
        wav = Wav()
        upload_init_message.wav.CopyFrom(wav)

    upload_request.upload_init_message.CopyFrom(upload_init_message)
    yield upload_request

    while True:
        data = file_handle.read(max_chunk_size_bytes)
        if not data:
            log.info("Done reading data")
            break
        upload_request = UploadRequest()
        upload_request.data_chunk = data
        yield upload_request

def validate_resource_arguments(args):
    if args.type != "activeprompt" and args.type != "user_dictionary" and args.type != "text_ruleset" and args.type != "wav":
        log.error("Invalid resource type [%s]. Must be one of: [activeprompt, user_dictionary, text_ruleset, wav]" % args.type)
        return False

    if args.type == "activeprompt":
        if not args.voice:
            log.error("Missing voice for type=activeprompt")
            return False
        elif not args.voice_model:
            log.error("Missing voice_model for type=activeprompt")
            return False
        elif not args.voice_version:
            log.error("Missing voice_version for type=activeprompt")
            return False
        elif not args.vocalizer_studio_version:
            log.error("Missing vocalizer_studio_version for type=activeprompt")
            return False
    elif args.type == "user_dictionary" and not args.language:
        log.error("Missing language for type=user_dictionary")
        return False
    elif args.type == "text_ruleset" and not args.language:
        log.error("Missing language for type=text_ruleset")
        return False

    return True

def run():
    parse_args()

    log_level = logging.DEBUG
    global log
    log = logging.getLogger('')
    logging.basicConfig(format='%(asctime)s %(levelname)-5s %(message)s', level=log_level)

    if args.upload and args.delete:
        log.error("--upload and --delete are mutually exclusive: choose only one operation at a time")
        return

    if args.upload:

        if not args.file:
            log.error("Missing file")
            return
        elif not args.context_tag:
            log.error("Missing context tag")
            return
        elif not args.name:
            log.error("Missing resource name")
            return

        if not os.path.exists(args.file):
            log.error("File [%s] does not exist" % args.file)
            return

        if not os.path.isfile(args.file):
            log.error("File [%s] is not a file" % args.file)
            return

        if not validate_resource_arguments(args):
            return

        with create_channel() as channel:
            storage_stub = StorageStub(channel)
            request_iterator = read_file(file=args.file, context_tag=args.context_tag, name=args.name, type=args.type, voice=args.voice, voice_model=args.voice_model, voice_version=args.voice_version, vocalizer_studio_version=args.vocalizer_studio_version, language=args.language, max_chunk_size_bytes=args.max_chunk_size_bytes)
            upload_response = storage_stub.Upload(request_iterator)
            log.info(text_format.MessageToString(upload_response))
    elif args.delete:
        if not args.uri:
            log.error("Missing URI for deletion")
            return

        with create_channel() as channel:
            storage_stub = StorageStub(channel)
            delete_request = DeleteRequest()
            delete_request.uri = args.uri
            delete_response = storage_stub.Delete(delete_request)
            log.info(text_format.MessageToString(delete_response))
    else:
        log.error("Missing operation flag: must set either --upload or --delete")
        return


if __name__ == '__main__':
    run()